@@ -2,20 +2,18 @@ import math
 import random
 import os
 import numpy as np
-from neurograd import xp, Tensor
+from neurograd import xp, Tensor, float32, int64
 from typing import Optional, List, Tuple, Union, Callable
 import glob
 from collections import deque
 
-
-
 # Try to import DALI first - this determines our capabilities
 try:
     import nvidia.dali as dali
     from nvidia.dali import pipeline_def, Pipeline
     import nvidia.dali.fn as fn
     import nvidia.dali.types as types
-    from nvidia.dali.plugin.base_iterator import LastBatchPolicy
+    from nvidia.dali.plugin.pytorch import DALIGenericIterator
     DALI_AVAILABLE = True
 except ImportError:
     DALI_AVAILABLE = False
@@ -25,7 +23,6 @@ except ImportError:
 # Always import OpenCV as fallback
 try:
     import cv2
-    # cv2.setNumThreads(1)  # Uncomment if needed for threading control
     CV2_AVAILABLE = True
 except ImportError:
     CV2_AVAILABLE = False
@@ -79,9 +76,8 @@ class Dataset:
 
 class ImageFolder(Dataset):
     """
-    Enhanced ImageFolder that uses DALI when available, OpenCV as fallback.
-    
-    Automatically detects DALI availability and optimizes accordingly.
+    ImageFolder dataset class that handles image loading and preprocessing.
+    All data loading parameters are now handled by the DataLoader.
     """
     def __init__(
         self,
@@ -89,15 +85,11 @@ class ImageFolder(Dataset):
         img_shape: tuple = None,          # (H, W)
         img_mode: str = "RGB",            # "RGB", "L", etc.
         img_normalize: bool = True,       # /255 -> float
-        img_transform: Union[Pipeline, Callable, None] = None,  # DALI pipeline or callable
+        img_transform: callable = None,   # DALI pipeline or callable
         target_transform: callable = None,
         img_dtype=xp.float32,
         target_dtype=xp.int64,
         chw: bool = True,                 # return CxHxW if True, else HxWxC
-        device: str = "cpu",              # "cpu" or "gpu"/"cuda" (DALI only)
-        num_threads: int = 4,             # DALI CPU threads
-        prefetch_queue_depth: int = 2,    # DALI prefetch depth
-        seed: int = 42
     ):
         self.root = root
         self.img_shape = img_shape
@@ -108,21 +100,15 @@ class ImageFolder(Dataset):
         self.img_dtype = img_dtype
         self.target_dtype = target_dtype
         self.chw = chw
-        self.device = "cuda" if device in ["gpu", "cuda"] else "cpu"
-        self.num_threads = num_threads
-        self.prefetch_queue_depth = prefetch_queue_depth
-        self.seed = seed
-
-        # Determine which backend to use
-        self.use_dali = DALI_AVAILABLE and isinstance(img_transform, (Pipeline, type(None)))
-        if device in ["gpu", "cuda"] and not DALI_AVAILABLE:
-            print("WARNING: GPU device requested but DALI not available. Using CPU with OpenCV fallback.")
-            self.device = "cpu"
 
         self.images: List[str] = []
         self.targets: List[str] = []
         self._collect_paths()
 
+        # Check if we have any images
+        if len(self.images) == 0:
+            raise ValueError(f"No images found in {root} with supported extensions: {IMG_EXTS}")
+
         # Stable class mapping
         self.target_names = sorted(set(self.targets))
         self.target_mapping = {name: i for i, name in enumerate(self.target_names)}
@@ -132,39 +118,46 @@ class ImageFolder(Dataset):
         self.numeric_targets = [self.target_mapping[t] for t in self.targets]
 
         print(f"ImageFolder initialized: {len(self)} samples, {self.num_classes} classes")
-        print(f"Backend: {'DALI' if self.use_dali else 'OpenCV'}, Device: {self.device}")
 
     def _collect_paths(self):
         """Collect image paths and their class labels"""
-        if os.path.exists(self.root) and os.path.isdir(self.root):
-            # Method 1: Class folders (ImageNet style)
-            for class_name in os.listdir(self.root):
-                class_path = os.path.join(self.root, class_name)
-                if not os.path.isdir(class_path):
-                    continue
-                
-                for ext in IMG_EXTS:
-                    pattern = os.path.join(class_path, f"*{ext}")
-                    for img_path in glob.glob(pattern):
-                        self.images.append(img_path)
-                        self.targets.append(class_name)
+        if not os.path.exists(self.root) or not os.path.isdir(self.root):
+            raise ValueError(f"Root directory {self.root} does not exist or is not a directory")
+        
+        # Method 1: Class folders (ImageNet style)
+        for class_name in os.listdir(self.root):
+            class_path = os.path.join(self.root, class_name)
+            if not os.path.isdir(class_path):
+                continue
             
-            # Method 2: If no class folders found, walk directory tree
-            if not self.images:
-                for r, _, files in os.walk(self.root):
-                    for f in files:
-                        if f.lower().endswith(IMG_EXTS):
-                            p = os.path.join(r, f)
-                            cls = os.path.basename(os.path.dirname(p))
-                            self.images.append(p)
-                            self.targets.append(cls)
+            for ext in IMG_EXTS:
+                pattern = os.path.join(class_path, f"*{ext}")
+                for img_path in glob.glob(pattern):
+                    self.images.append(img_path)
+                    self.targets.append(class_name)
+        
+        # Method 2: If no class folders found, walk directory tree
+        if not self.images:
+            for r, _, files in os.walk(self.root):
+                for f in files:
+                    if any(f.lower().endswith(ext) for ext in IMG_EXTS):
+                        p = os.path.join(r, f)
+                        cls = os.path.basename(os.path.dirname(p))
+                        self.images.append(p)
+                        self.targets.append(cls)
 
-    def get_dali_pipeline(self, batch_size: int, shuffle: bool = True) -> Optional[Pipeline]:
+    def get_dali_pipeline(self, batch_size: int, shuffle: bool = True, 
+                         device: str = "cpu", num_threads: int = 4, 
+                         seed: int = 42):
         """Create DALI pipeline for this dataset (only when DALI is available)"""
-        if not self.use_dali:
+        if not DALI_AVAILABLE:
             return None
             
-        @pipeline_def(batch_size=batch_size, num_threads=self.num_threads, device_id=0, seed=self.seed)
+        # If user provided an entire DALI Pipeline instance, return it directly.
+        if isinstance(self.img_transform, Pipeline):
+            return self.img_transform
+
+        @pipeline_def(batch_size=batch_size, num_threads=num_threads, device_id=0, seed=seed)
         def image_pipeline():
             # Create file reader
             if shuffle:
@@ -172,7 +165,7 @@ class ImageFolder(Dataset):
                     files=self.images,
                     labels=self.numeric_targets,
                     random_shuffle=True,
-                    initial_fill=batch_size * 4,  # Buffer size
+                    initial_fill=batch_size * 4,
                     name="Reader"
                 )
             else:
@@ -183,56 +176,50 @@ class ImageFolder(Dataset):
                     name="Reader"
                 )
 
-            # Decode images
-            if self.device == "cuda":
-                # Use mixed precision decoding (CPU decode, GPU processing)
-                images = fn.decoders.image(images, device="mixed", output_type=types.RGB)
-            else:
-                images = fn.decoders.image(images, device="cpu", output_type=types.RGB)
+            # Decode and apply RandomResizedCrop
+            h, w = self.img_shape if self.img_shape is not None else (224, 224)
+            images = fn.decoders.image_random_crop(
+                images,
+                device=device,
+                output_type=types.RGB,
+                random_aspect_ratio=[0.8, 1.25],
+                random_area=[0.1, 1.0],
+                num_attempts=100
+            )
+            
+            # Resize the cropped image to the final desired shape
+            images = fn.resize(
+                images,
+                device=device,
+                size=[h, w],
+                interp_type=types.INTERP_LINEAR
+            )
 
-            # Resize if needed
-            if self.img_shape is not None:
-                h, w = self.img_shape
-                images = fn.resize(
-                    images,
-                    device=self.device,
-                    size=[h, w],
-                    interp_type=types.INTERP_LINEAR
-                )
-
-            # Handle grayscale conversion
+            # Handle grayscale conversion if needed
             if self.img_mode.upper() in ("L", "GRAY", "GREY", "GRAYSCALE"):
-                # Convert to grayscale
                 images = fn.color_space_conversion(
-                    images,
-                    device=self.device,
-                    image_type=types.RGB,
-                    output_type=types.GRAY
+                    images, device=device, image_type=types.RGB, output_type=types.GRAY
                 )
 
-            # Apply custom transforms if provided
-            if self.img_transform is not None:
-                if isinstance(self.img_transform, Pipeline):
-                    # If it's already a DALI pipeline, we'd need to integrate it
-                    # This is complex, so for now we'll apply it post-DALI
-                    pass
+            # Apply remaining augmentations
+            if self.img_transform and callable(self.img_transform):
+                try:
+                    images = self.img_transform(images)
+                except Exception as e:
+                    print(f"WARNING: Failed to apply DALI callable transform. Error: {e}. Skipping transform.")
 
             # Normalize to [0, 1] if requested
             if self.img_normalize:
-                images = fn.cast(images, device=self.device, dtype=types.FLOAT)
+                images = fn.cast(images, dtype=types.FLOAT)
                 images = images / 255.0
             else:
-                images = fn.cast(images, device=self.device, dtype=types.FLOAT)
+                images = fn.cast(images, dtype=types.FLOAT)
 
-            # Convert to CHW format if requested
-            if not self.chw:
-                # DALI outputs CHW by default after transpose, so we need to handle HWC
-                if len(images.shape) == 4:  # Batch dimension
-                    images = fn.transpose(images, device=self.device, perm=[0, 2, 3, 1])
-                else:
-                    images = fn.transpose(images, device=self.device, perm=[1, 2, 0])
-
-            return images, labels.gpu() if self.device == "cuda" else labels
+            # Transpose to CHW format if requested
+            if self.chw:
+                images = fn.transpose(images, device=device, perm=[2, 0, 1])
+            
+            return images, labels.gpu() if device == "gpu" else labels
 
         return image_pipeline()
 
@@ -240,10 +227,10 @@ class ImageFolder(Dataset):
         """Apply image transforms (OpenCV fallback path)"""
         if self.img_transform is None:
             return arr
-        
-        # Skip DALI pipelines in fallback mode
-        if DALI_AVAILABLE and isinstance(self.img_transform, Pipeline):
-            print("WARNING: DALI Pipeline transform provided but using fallback mode. Transform skipped.")
+            
+        if isinstance(self.img_transform, Pipeline):
+            print("WARNING: DALI Pipeline transform provided but running in fallback numpy/OpenCV mode. "
+                  "Cannot apply Pipeline to ndarray; skipping transform.")
             return arr
             
         # Try Albumentations-style call
@@ -254,9 +241,13 @@ class ImageFolder(Dataset):
             return out
         except TypeError:
             pass
-        
+
         # Fallback: plain callable expecting ndarray
-        return self.img_transform(arr)
+        try:
+            return self.img_transform(arr)
+        except Exception as e:
+            print(f"WARNING: img_transform callable raised an exception: {e}. Returning original image.")
+            return arr
 
     def _load_image_opencv(self, path: str) -> np.ndarray:
         """Load image using OpenCV (fallback implementation)"""
@@ -268,11 +259,10 @@ class ImageFolder(Dataset):
         if mode in ("L", "GRAY", "GREY", "GRAYSCALE"):
             flag = cv2.IMREAD_GRAYSCALE
         elif mode == "RGBA":
-            flag = cv2.IMREAD_UNCHANGED  # preserve alpha if present
+            flag = cv2.IMREAD_UNCHANGED
         else:
-            flag = cv2.IMREAD_COLOR  # BGR
+            flag = cv2.IMREAD_COLOR
         
-        # Avoid EXIF orientation work
         try:
             flag |= cv2.IMREAD_IGNORE_ORIENTATION
         except Exception:
@@ -288,7 +278,7 @@ class ImageFolder(Dataset):
         elif mode == "RGBA" and arr.ndim == 3 and arr.shape[2] == 4:
             arr = cv2.cvtColor(arr, cv2.COLOR_BGRA2RGBA)
 
-        # Resize if requested (cv2 expects (W,H))
+        # Resize if requested
         if self.img_shape is not None:
             h, w = self.img_shape
             arr = cv2.resize(arr, (int(w), int(h)), interpolation=cv2.INTER_LINEAR)
@@ -302,7 +292,7 @@ class ImageFolder(Dataset):
         
         # Convert to CHW if requested
         if self.chw and arr.ndim == 3:
-            arr = np.transpose(arr, (2, 0, 1))  # C,H,W
+            arr = np.transpose(arr, (2, 0, 1))
         
         # Normalize
         if self.img_normalize:
@@ -313,18 +303,12 @@ class ImageFolder(Dataset):
         return arr
 
     def __getitem__(self, idx: int):
-        """Get single item - uses appropriate backend"""
+        """Get single item - uses OpenCV fallback"""
         img_path = self.images[idx]
         target = self.numeric_targets[idx]
         
-        # Load image using appropriate backend
-        if self.use_dali:
-            # For DALI datasets, single item access uses OpenCV fallback
-            # DALI is optimized for batch processing
-            image = self._load_image_opencv(img_path)
-        else:
-            # Standard OpenCV loading
-            image = self._load_image_opencv(img_path)
+        # Load image using OpenCV
+        image = self._load_image_opencv(img_path)
         
         # Apply target transform
         if self.target_transform:
@@ -349,8 +333,11 @@ class ImageFolder(Dataset):
         for i in range(len(self)):
             yield self[i]
 
+    def __str__(self):
+        return (f"ImageFolder(root='{self.root}', samples={len(self)}, "
+                f"classes={self.num_classes})")
+
     def __repr__(self):
-        backend = "DALI" if self.use_dali else "OpenCV"
         shape = None
         if len(self) > 0:
             try:
@@ -360,56 +347,45 @@ class ImageFolder(Dataset):
                 shape = None
         
         return (f"ImageFolder(root='{self.root}', samples={len(self)}, "
-                f"classes={self.num_classes}, backend={backend}, device='{self.device}', "
+                f"classes={self.num_classes}, "
                 f"shape={shape}, img_dtype={self.img_dtype}, target_dtype={self.target_dtype}, "
                 f"mode='{self.img_mode}', normalize={self.img_normalize}, chw={self.chw})")
 
 
 class DataLoader:
     """
-    Enhanced DataLoader with DALI acceleration when available, OpenCV fallback otherwise.
-    
-    Automatically detects capabilities and optimizes accordingly.
+    DataLoader that handles all data loading parameters including device, threading, and prefetching.
+    Automatically uses DALI when available and appropriate, with OpenCV fallback.
     """
     def __init__(
         self,
         dataset: Union[ImageFolder, Dataset],
         batch_size: int = 32,
         shuffle: bool = True,
-        num_threads: int = 4,
         device: str = "cpu",
-        prefetch_queue_depth: int = 2,
+        num_workers: int = 4,
+        prefetch_batches: int = 2,
         drop_last: bool = False,
         seed: Optional[int] = None,
-        num_workers: Optional[int] = None,  # For fallback threading
-        prefetch_batches: int = 2           # For fallback prefetch
     ):
         self.dataset = dataset
         self.batch_size = batch_size
         self.shuffle = shuffle
-        self.num_threads = num_threads
-        self.device = "cuda" if device in ["gpu", "cuda"] else "cpu"
-        self.prefetch_queue_depth = prefetch_queue_depth
+        self.device = "gpu" if device in ["gpu", "cuda"] else "cpu"
+        self.num_workers = max(0, int(num_workers))
+        self.prefetch_batches = max(0, int(prefetch_batches))
         self.drop_last = drop_last
         self.seed = seed or 42
-        self.prefetch_batches = max(0, int(prefetch_batches))
-        
-        # Determine threading for fallback mode
-        if num_workers is None:
-            cores = os.cpu_count() or 2
-            self.num_workers = max(1, min(8, cores - 1))
-        else:
-            self.num_workers = int(num_workers)
         
         self._pipeline = None
         self._dali_iter = None
         self._executor: Optional[ThreadPoolExecutor] = None
         
-        # Initialize DALI pipeline if available and applicable
+        # Determine if we should use DALI
         self.use_dali = (
             DALI_AVAILABLE and 
             isinstance(dataset, ImageFolder) and 
-            dataset.use_dali
+            self.device in ["gpu", "cuda"]
         )
         
         if self.use_dali:
@@ -425,22 +401,29 @@ class DataLoader:
         if not self.use_dali or not isinstance(self.dataset, ImageFolder):
             return
         
-        # Update dataset's DALI settings
-        self.dataset.device = self.device
-        self.dataset.num_threads = self.num_threads
-        self.dataset.prefetch_queue_depth = self.prefetch_queue_depth
-        self.dataset.seed = self.seed
-        
         # Create pipeline
         self._pipeline = self.dataset.get_dali_pipeline(
             batch_size=self.batch_size,
-            shuffle=self.shuffle
+            shuffle=self.shuffle,
+            device=self.device,
+            num_threads=self.num_workers,
+            seed=self.seed
         )
         
         if self._pipeline is not None:
             # Build the pipeline
             self._pipeline.build()
-            print(f"DALI pipeline initialized: batch_size={self.batch_size}, device={self.device}")
+            
+            # Create DALI iterator
+            last_batch_policy = 'drop' if self.drop_last else 'part'
+            self._dali_iter = DALIGenericIterator(
+                self._pipeline, 
+                output_map=["images", "labels"],
+                last_batch_policy=last_batch_policy,
+                auto_reset=True
+            )
+            print(f"DALI pipeline initialized: batch_size={self.batch_size}, device={self.device}, "
+                  f"num_threads={self.num_workers}")
 
     def __len__(self):
         n = len(self.dataset)
@@ -450,67 +433,29 @@ class DataLoader:
 
     def __iter__(self):
         """Main iteration - uses DALI when available, falls back otherwise"""
-        if self.use_dali and self._pipeline is not None:
+        if self.use_dali and self._dali_iter is not None:
             return self._dali_iterator()
         else:
             return self._regular_iterator()
 
     def _dali_iterator(self):
         """DALI-based iteration - maximum performance!"""
-        # Reset pipeline for new epoch
-        self._pipeline.reset()
-        
-        # Create DALI iterator
-        last_batch_policy = LastBatchPolicy.DROP if self.drop_last else LastBatchPolicy.FILL
-        
-        class DALIGenericIterator:
-            def __init__(self, pipeline, output_map, size, last_batch_policy):
-                self.pipeline = pipeline
-                self.output_map = output_map
-                self.size = size
-                self.last_batch_policy = last_batch_policy
-                self._counter = 0
+        for data in self._dali_iter:
+            # Extract data from DALI output
+            images = data[0]["images"]
+            labels = data[0]["labels"]
             
-            def __iter__(self):
-                return self
+            # Convert to neurograd tensors
+            if self.device == "gpu":
+                # Use DLPack for zero-copy conversion
+                X = Tensor(xp.from_dlpack(images), dtype=self.dataset.img_dtype)
+                y = Tensor(xp.from_dlpack(labels), dtype=self.dataset.target_dtype)
+            else:
+                # Convert to numpy arrays
+                X = Tensor(images.numpy(), dtype=self.dataset.img_dtype)
+                y = Tensor(labels.numpy(), dtype=self.dataset.target_dtype)
             
-            def __next__(self):
-                if self._counter >= self.size:
-                    raise StopIteration
-                
-                try:
-                    outputs = self.pipeline.run()
-                    self._counter += 1
-                    
-                    # Extract images and labels
-                    images = outputs[0].as_array() if hasattr(outputs[0], 'as_array') else outputs[0]
-                    labels = outputs[1].as_array() if hasattr(outputs[1], 'as_array') else outputs[1]
-                    
-                    # Convert to CPU if on GPU
-                    if self.pipeline.device == "cuda":
-                        if hasattr(images, 'as_cpu'):
-                            images = images.as_cpu().as_array()
-                        if hasattr(labels, 'as_cpu'):
-                            labels = labels.as_cpu().as_array()
-                    
-                    # Convert to Tensors
-                    X = Tensor(images, dtype=self.dataset.img_dtype)
-                    y = Tensor(labels, dtype=self.dataset.target_dtype)
-                    
-                    return X, y
-                    
-                except StopIteration:
-                    raise StopIteration
-        
-        iterator = DALIGenericIterator(
-            pipeline=self._pipeline,
-            output_map=["images", "labels"],
-            size=len(self),
-            last_batch_policy=last_batch_policy
-        )
-        
-        for batch in iterator:
-            yield batch
+            yield X, y
 
     def _regular_iterator(self):
         """Regular iteration for non-DALI datasets with threading and prefetch"""
@@ -573,7 +518,7 @@ class DataLoader:
             return [self._executor.submit(self.dataset.__getitem__, i) for i in idxs]
         else:
             # synchronous path for num_workers=0
-            return [(self.dataset[i], None) for i in idxs]  # (result, None) to unify interface
+            return [(self.dataset[i], None) for i in idxs]
 
     def _gather_batch(self, futures_or_results):
         """Block until the batch is ready, then stack into (X, y) Tensors"""
@@ -618,11 +563,10 @@ class DataLoader:
 
     def __repr__(self):
         backend = "DALI" if self.use_dali else "OpenCV+Threading"
-        device_str = f", device='{self.device}'" if self.use_dali else ""
         return (f"<DataLoader: {len(self)} batches, "
                 f"batch_size={self.batch_size}, shuffle={self.shuffle}, "
-                f"backend={backend}{device_str}, "
-                f"prefetch_depth={self.prefetch_queue_depth if self.use_dali else self.prefetch_batches}>")
+                f"backend={backend}, device='{self.device}', "
+                f"num_workers={self.num_workers}, prefetch_batches={self.prefetch_batches}>")
 
 
 # DALI transform creation helper (only available when DALI is installed)
@@ -668,46 +612,10 @@ def create_dali_transforms(
                 images,
                 device=device,
                 angle=fn.random.uniform(range=[-rotation_angle, rotation_angle]),
+                keep_size=True,
                 fill_value=0
             )
         
         return images
     
-    return apply_transforms
-
-
-# Example usage demonstrating graceful fallback:
-"""
-# Basic usage - works with or without DALI
-dataset = ImageFolder(
-    root="/path/to/imagenet",
-    img_shape=(224, 224),
-    img_mode="RGB", 
-    img_normalize=True,
-    device="cuda" if DALI_AVAILABLE else "cpu",  # Automatically adapts
-    num_threads=8,
-    prefetch_queue_depth=3
-)
-
-dataloader = DataLoader(
-    dataset,
-    batch_size=128,
-    shuffle=True,
-    device="cuda" if DALI_AVAILABLE else "cpu",  # Automatically adapts
-    prefetch_queue_depth=3,
-    num_threads=8
-)
-
-print(f"Using backend: {'DALI' if dataloader.use_dali else 'OpenCV+Threading'}")
-
-# Training loop - works at maximum speed regardless of backend
-for epoch in range(num_epochs):
-    for batch_idx, (images, targets) in enumerate(dataloader):
-        # Your training code here
-        # Performance automatically optimized based on available libraries
-        pass
-    dataloader.reset()  # Reset for next epoch
-    
-# Clean up when done
-dataloader.close()
-"""
\ No newline at end of file
+    return apply_transforms
\ No newline at end of file
